{$S-}
uses
   Objects,
   Compiler;

var
   EC              :TExpressionCompiler;
   Code            :TDosStream;

   i               :Integer;

function  OpSgn(x                      :Integer) :Integer; far;
begin
   if x=0 then OpSgn:=0 else
   if x>0 then OpSgn:=1 else OpSgn:=-1;
end;

function  OpF(x                        :Integer) :Integer; far;
assembler;
asm
   mov  cx,x
   cmp  cx,7
   ja   @@Overflow
   mov  ax,1
   or   cx,cx
   jz   @@Exit
@@Loop:
   mul  cx
   loop @@Loop
@@Exit:
   pop  bp
   retf 2
@@Overflow:
   xor  ax,ax
end;

var
   Vars            :Array [1..4] of Integer;
   CodeF           :function :Integer;

var
   Timer           :Byte absolute $0040:$006C;
   LTimer          :Byte;
   n1, n2          :LongInt;

function  TestF    :Integer;
begin
   TestF:=(Vars[3] * 5 + 6 * Vars[4]) + OpF(Vars[1]) + 2 * OpSgn(2+Vars[2])
end;

var
   Expr            :String;
   CodeR           :function :Real;

begin
   EC.Init;

   {EC.CaseSensitive:=True;}
   {EC.Optimization:=False;}
   {EC.SafeMode:=True;}
   {Test8086:=0;}

   EC.DefineAll;

   EC.AddOp('sgn', tiNone, tiInteger, tiInteger, 30, @OpSgn);
   EC.AddOp('!', tiInteger, tiNone, tiInteger, 50, @OpF);

   EC.AddVar('Var1', Vars[1], tiInteger);
   EC.AddVar('Var2', Vars[2], tiInteger);
   EC.AddVar('Var3', Vars[3], tiInteger);
   EC.AddVar('Var4', Vars[4], tiInteger);

   { The main testing expression }
   EC.Compile('(Var3 * (1+1 +1*1*(1*1*(1*1)*1)*(1*1*1)+ 2) + 6 * Var4) +Var1! + 2 * sgn(2 + Var2)', tiInteger);
{  EC.Compile('(var3 * (1+1 +1*1*(1*1*(1*1)*1)*(1*1*1)+ 2) + 6 * VAR4) +Var1! + 2 * SGN(2 + Var2)', tiInteger);}

   { Real constants }
{  EC.Compile('Integer(1.0+2+3+4+5+6+7+8+9+10+11+12+13+14+15)', tiInteger);}

   { Hexadecimal formats }
{  EC.Compile('0xF0 or $0F or 0F00h or 0FF80h', tiInteger);}
{  EC.Compile('0Xf0 or $0f or 0f00H or 0ff80h', tiInteger);}

   { Bitwise operations }
{  EC.Compile('(16 shr 2) or 8', tiInteger);}

   { Real E-format }
{  EC.Compile('Integer(1E-3*1E4)', tiInteger);}

   { Accessing variables }
{  EC.Compile('Var1', tiInteger);}
{  EC.Compile('Var1*Var2*Var3', tiInteger);}
{  EC.Compile('Integer(Var1^(Var1^(1/2)))', tiInteger);}
{  EC.Compile('Integer(Var1- - Var1)', tiInteger);}
{  EC.Compile('Integer(3.0- -2.0)', tiInteger);}

   { Operations must be separated with spaces }
{  EC.Compile('-1-2* - - sgn -(1-10)', tiInteger);}
{  EC.Compile('1+ -(1-10)', tiInteger);}
{  EC.Compile('Var1- -Var1', tiInteger);}
{  EC.Compile('Var1! +2', tiInteger);}

   { Just a division test }
{  EC.Compile('Integer((-10 div -3)=(10 div 3))', tiInteger);}

   { Real functions }
{  EC.Compile('Integer(1000.*arcsin 0.5)', tiInteger);}
{  EC.Compile('Integer(1000.*arccos -0.5)', tiInteger);}

   { There can be two kinds of unary operation }
{  EC.Compile('!$FF00', tiInteger);}
{  EC.Compile('5!', tiInteger);}

   { Integer <-> Boolean <-> Real }
{  EC.Compile('-8 < -2', 'Boolean');}
{  EC.Compile('Integer((true or false) and (8>5))', tiInteger);}
{  EC.Compile('Integer(2.51>2.5)', tiInteger);}
{  EC.Compile('Integer(True>False)*10+Integer(True<False)', tiInteger);}
{  EC.Compile('Integer(3>=4)', tiInteger);}

   { Type adjustment }
{  EC.Compile('Integer(Var1^3)', tiInteger);}
{  EC.Compile('Integer((1+2)^(6 div 2))', tiInteger);}
{  EC.Compile('Integer(1+2.0)', tiInteger);}
{  EC.Compile('Integer((Var4-Var1)^(Var4 div Var1))', tiInteger);}

   { Result type mismatch (ecInvType) }
{  EC.Compile('6.0/2.0', tiInteger);}

{  EC.Compile('Var1 - (Var1+1)', tiInteger);}

   WriteLn(#13#10'Status is ', EC.Status,
      ', code size is ', EC.CodeSize, ' bytes'#13#10);

   Code.Init('CODE.DAT', stCreate);
   Code.Write(EC.Code^, EC.CodeSize);
   Code.Done;

   Pointer(@CodeF):=EC.Code;

   if EC.Status<>ecOk then
   begin
      EC.Done; Halt;
   end;

   Vars[1]:=5; Vars[2]:=-3; Vars[3]:=46; Vars[4]:=9;

   WriteLn('TestF = ', TestF);
   WriteLn('CodeF = ', CodeF);

   for i:=Low(Vars) to High(Vars) do Inc(Vars[i], i);

   WriteLn('TestF = ', TestF);
   WriteLn('CodeF = ', CodeF);

   n1:=0;
   LTimer:=Timer;
   while Timer=LTimer do;
   LTimer:=Timer;
   while Timer=LTimer do
   begin
      i:=TestF;
      Inc(n1);
   end;

   n2:=0;
   LTimer:=Timer;
   while Timer=LTimer do;
   LTimer:=Timer;
   while Timer=LTimer do
   begin
      i:=CodeF;
      Inc(n2);
   end;

   WriteLn((1/n2-1/n1)*n2*100:3:2, ' % speed decrease');

   WriteLn(#13#10'Enter an expression to evaluate (the result must be real) :');
   ReadLn(Expr);

   EC.Compile(Expr, tiReal);
   Pointer(@CodeR):=EC.Code;

   if EC.Status<>ecOk
   then WriteLn('Error #', EC.Status)
   else WriteLn('The result is ', CodeR:3:3);

   EC.Done;
end.


